package automaton;
import heap.*;
import code.*;
import java.util.*;
import util.*;

public class Automaton {
	private List states;
	private List transitions;
	private List counters;
	private State first;
	private State error;
	
	public Automaton()
	{
		states = null;
		transitions = null;
		first = null;
		error = null;
		counters = null;
	}
	
	
	public void generate(Heap h,CodeNode p,List integerslist)
	{
		/* assume that h is normalized */
		
		/* init states table */
		int states_c = 1;
		states = new LinkedList();
		transitions = new LinkedList();
		counters = new LinkedList(integerslist);
		first = new State(h,false,(Statement)p);
		states.add(first); // first state
		this.addCounters(h);
		first.setId(states_c);
		states_c++;
		/* generation */
		LinkedList state_list = new LinkedList();
		EquivalenceList mark = new EquivalenceList();
		state_list.add(first);		
		while (!state_list.isEmpty()) {
				//main.Main.getOptions().Debug(this);
				State s = (State) state_list.getFirst();
				mark.add(s);
				state_list.remove(s);					
				if (!s.isFinal()) {
					Map succ = s.next();
					Set e = succ.entrySet();
					Iterator j = e.iterator();
					while (j.hasNext()) {
						Map.Entry tmp = (Map.Entry) j.next();
						State r = (State) tmp.getKey();
						Transition t = (Transition) tmp.getValue();
						if(!mark.contains(r))
						{
							transitions.add(t);
							if(r.isError())
							{
								if(!states.contains(r))
								{
									states.add(r);
									error = r;
									mark.add(r);
								}
							}
							else if(!state_list.contains(r))
							{
								r.setId(states_c);
								states_c++;
								state_list.addLast(r);
								states.add(r);
								this.addCounters(r.getHeap());
							}
						}
						else
						{
							if(r.isError())
							{
								t.setSfinal(error);
								if(!transitions.contains(t))
									transitions.add(t);
							}
							else
							{
								states.add(r);
								this.addCounters(r.getHeap());
								r.setId(states_c);
								states_c++;
								if(!transitions.contains(t))
									transitions.add(t);
								
								State o = (State) mark.getFirstEquivalent(r);
								Transition u = new Transition(r,o,CodeNode.voidGuard(),CodeNode.voidActions());
								u.heq((Map)o.getHeap().isEquivalentTo(r.getHeap()));								
								if(!transitions.contains(u))
									transitions.add(u);
							}
							
						}
					}
				}				
			}		
	}
	
	public List getCounters(){
		return counters;
	}
	
	public State getFirst() {
		return first;
	}

	public List getStates() {
		return states;
	}

	public List getTransitions() {
		return transitions;
	}
	
	public State get(State r)
	{
		Iterator i = states.iterator();
		while(i.hasNext())
		{
			State s = (State)i.next();
			if(s.equals(r))
				return s;
		}
		return null;
	}
	
	private void addCounters(Heap h){
		Iterator i = h.getCounters().iterator();
		while(i.hasNext())
		{
			String s = (String)i.next();
			if(!counters.contains(s))
				counters.add(s);
		}
	}
	
	/**Minimizes this automaton. Consecutives nodes with same heap are merged. */
	// TODO : final states minimization
	public void minimize()
	{
		ListIterator i = states.listIterator(states.size());
		while(i.hasPrevious())
		{
			State s = (State)i.previous();
			List succs = successors(s);
			List preds = predecessors(s);
			if(succs.size() == 1 && preds.size() == 1)
			{
				State succ = (State)succs.get(0);
				State pred = (State)preds.get(0);
				
				if(pred.getHeap().equals(s.getHeap()))
				{
					Transition before = getTransition(pred, s);
					Transition after = getTransition(s, succ);
					
					// merge transitions
					Transition t = new Transition(pred,succ,CodeNode.and(before.getGuard(), after.getGuard()),CodeNode.slist(before.getActions(), after.getActions()));
					transitions.add(t);
						
						
					// remove states and transitions
					transitions.remove(before);
					transitions.remove(after);
						
					i.remove();					
				}
			}
		}
	}
	
	/**Builds the list of all successors of a state in this automaton. */
	private List successors(State s)
	{
		if(s == null)
			return null;
		
		List l = new LinkedList();
		Iterator i = transitions.iterator();
		while(i.hasNext())
		{
			Transition t = (Transition)i.next();
			if(s.equals(t.getSinitial()))
				l.add(t.getSfinal());
		}
		
		return l;
	}
	
	/**Builds the list of all predecessors of a state in this automaton. */
	private List predecessors(State s)
	{
		if(s == null)
			return null;
		
		List l = new LinkedList();
		Iterator i = transitions.iterator();
		while(i.hasNext())
		{
			Transition t = (Transition)i.next();
			if(s.equals(t.getSfinal()))
				l.add(t.getSinitial());
		}
		
		return l;
	}
	
	/**Return the transition from one state to another in this automaton */
	private Transition getTransition(State from,State to)
	{
		if( from == null || to == null)
			return null;
		
		Iterator i = transitions.iterator();
		while(i.hasNext())
		{
			Transition t = (Transition)i.next();
			if(from.equals(t.getSinitial()) && to.equals(t.getSfinal()))
				return t;
		}
		return null;
	}
	
}
