package automaton;
import heap.*;
import code.*;
import main.*;
import util.*;
import java.util.*;

public class State implements Equivalent {
	public static final int ERROR_STATE = -1337; 
	private Heap heap;
	private boolean isFinal;
	private Statement instruction;
	private int id;
	
	public State(Heap h, boolean b,Statement p)
	{
		heap = h;
		isFinal = b;
		instruction = p;
		id = 0;
	}
	
	public State()
	{
		heap = null;
		isFinal = false;
		instruction = null;
		id = 0;
	}

	public Heap getHeap() {
		return heap;
	}

	public int getId(){
		return id;
	}
	
	public void setId(int i){
		id = i;
	}
	
	public void setHeap(Heap heap) {
		this.heap = heap;
	}

	public boolean isFinal() {
		return isFinal;
	}
	
	public boolean isError() {
		return id == ERROR_STATE;
	}
	
	public void setError() {
		heap = null;
		isFinal = false;
		instruction = null;
		id = ERROR_STATE;
	}

	public void setFinal(boolean isFinal) {
		this.isFinal = isFinal;
	}

	public Statement getInstruction() {
		return instruction;
	}

	public void setInstruction(Statement instruction) {
		this.instruction = instruction;
	}
	
	
	public Map next()
	{
		Map m = new HashMap();
		State s;
		Transition t;
		boolean pointerGuards = Main.getOptions().pointerGuards();
		CodeNode n,guard,action,action2;
		String x,y;
		int i;
		/* build resulting heap of instruction, and make states according to it */
		int type = instruction.instruction(heap);
		switch(type)
		{
		case Statement.error:
			s = new State();
			s.setError();
			
			t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.voidActions());
			m.put(s,t);
			break;
		case Statement.asgn_xint:
			s = new State();
			s.heap = heap;
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,CodeNode.voidGuard(),(CodeNode)instruction);
			m.put(s, t);
			break;
		case Statement.asgn_xnew:
			s = new State();
			s.heap = (Heap)heap.clone();
			s.heap.asignXToNew(((CNAsgn)instruction).leftId());
			n = s.heap.normalize();
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			x = s.heap.getVertex(new Root(((CNAsgn)instruction).leftId())).counter(); // [added, Marius]
			// t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.slist(n, CodeNode.actions_x1(((CNAsgn)instruction).leftId())));
			t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.slist(n, CodeNode.actions_x1(x)));
			m.put(s, t);			
			break;
		case Statement.asgn_xnull: 
			s = new State();
			s.heap = (Heap)heap.clone();
			s.heap.asignXToNull(((CNAsgn)instruction).leftId());
			n = s.heap.normalize();
			
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			// t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.slist(n, CodeNode.voidActions()));
			t = new Transition(this,s,CodeNode.voidGuard(), n);
			m.put(s, t);	
			break;
		case Statement.asgn_xx: 
			s = new State();
			s.heap = heap;
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
		
			t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.voidActions());
			m.put(s, t);
			break;
		case Statement.asgn_xy: 
			s = new State();
			s.heap = (Heap)heap.clone();
			s.heap.asignXToY(((CNAsgn)instruction).leftId(),((CNAsgn)instruction).rightId());
			n = s.heap.normalize();
			
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			// t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.slist(n, CodeNode.voidActions()));
			t = new Transition(this,s,CodeNode.voidGuard(), n);
			m.put(s, t);			
			break;
		case Statement.asgn_xxnext: 
			/* 2 etats : cx = 1 && cx > 1 */
			// cx = 1
			s = new State();
			s.heap = (Heap)heap.clone();
			x = s.heap.getVertex(new Root(((CNAsgn)instruction).leftId())).counter();
			guard = CodeNode.guard_x1(x);
			s.heap.asignXToXNextWithCounterXEquals1(((CNAsgn)instruction).leftId());
			n = s.heap.normalize();
			
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,guard,CodeNode.slist(n, CodeNode.voidActions()));
			m.put(s, t);	
			
			// cx > 1
			s = new State();
			s.heap = (Heap)heap.clone();
			guard = CodeNode.guard_x(x);
			action = CodeNode.actions_x1(x);
			s.heap.asignXToXNextWithCounterXGreaterThan1(((CNAsgn)instruction).leftId());
			y =  s.heap.getVertex(new Root(((CNAsgn)instruction).leftId())).counter();
			action2 = CodeNode.actions_xy_1(y, x);			
			action = CodeNode.slist(action2, action);
			n = s.heap.normalize();
			n = CodeNode.slist(action,n);
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,guard,n);
			m.put(s, t);			
			break;
		case Statement.asgn_xynext: 
			/* 2 etats : cy = 1 && cy > 1 */
			// cy = 1
			s = new State();
			s.heap = (Heap)heap.clone();
			y = s.heap.getVertex(new Root(((CNAsgn)instruction).rightId())).counter();
			guard = CodeNode.guard_x1(y);
			s.heap.asignXToYNextWithCounterYEquals1(((CNAsgn)instruction).leftId(),((CNAsgn)instruction).rightId());
			n = s.heap.normalize();
			
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			// t = new Transition(this,s,guard,CodeNode.slist(n, CodeNode.voidActions()));
			t = new Transition(this,s,guard,n);
			m.put(s, t);	
			
			// cy > 1
			s = new State();
			s.heap = (Heap)heap.clone();
			guard = CodeNode.guard_x(y);
			action = CodeNode.actions_x1(y);
			s.heap.asignXToYNextWithCounterYGreaterThan1(((CNAsgn)instruction).leftId(),((CNAsgn)instruction).rightId());
			x =  s.heap.getVertex(new Root(((CNAsgn)instruction).leftId())).counter();
			y =  s.heap.getVertex(new Root(((CNAsgn)instruction).rightId())).counter();
			action2 = CodeNode.actions_xy_1(x, y);			
			action = CodeNode.slist(action2, action);
			n = s.heap.normalize();
			n = CodeNode.slist(action,n);
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,guard,n);
			m.put(s, t);		
			break;
		case Statement.asgn_xnextnull: 
			s = new State();
			s.heap = (Heap)heap.clone();
			x = s.heap.getVertex(new Root(((CNAsgn)instruction).leftId())).counter();
			action = CodeNode.actions_x1(x);
			s.heap.asignXNextToNull(((CNAsgn)instruction).leftId());
			n = s.heap.normalize();
			
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.slist(n, action));
			m.put(s, t);			
			break;
		case Statement.asgn_xnexty: 
			s = new State();
			s.heap = (Heap)heap.clone();
			x = s.heap.getVertex(new Root(((CNAsgn)instruction).leftId())).counter();
			action = CodeNode.actions_x1(x);
			s.heap.asignXNextToY(((CNAsgn)instruction).leftId(),((CNAsgn)instruction).rightId());
			n = s.heap.normalize();
			
			
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.slist(n, action));
			m.put(s, t);			
			break;
		case Statement.if_: 
			i = instruction.next(heap);
			if(i == Statement.double_)
			{
				// 2 branchement possibles
				s = new State();
				s.heap = heap;
				s.instruction = instruction.next(true);
				s.isFinal = (s.instruction == null)?true:false;
				CodeNode a = (CodeNode)((CodeNode)instruction).jjtGetChild(0);
				t = new Transition(this,s,pointerGuards?a:CodeNode.voidGuard(),CodeNode.voidActions());
				m.put(s, t);
				
				s = new State();
				s.heap = heap;
				s.instruction = instruction.next(false);
				s.isFinal = (s.instruction == null)?true:false;
				CodeNode g = a.isVoid()?a:CodeNode.not(a);
				t = new Transition(this,s,pointerGuards?g:CodeNode.voidGuard(),CodeNode.voidActions());
				m.put(s, t);				
			}
			else
			{
				// un seul branchement possible
				s = new State();
				s.heap = heap;
				s.instruction = instruction.next((i == Statement.true_)?true:false);
				s.isFinal = (s.instruction == null)?true:false;
				CodeNode a = (CodeNode)((CodeNode)instruction).jjtGetChild(0);
				CodeNode g = (i == Statement.true_ || a.isVoid())?a:CodeNode.not(a);
				t = new Transition(this,s,pointerGuards?g:CodeNode.voidGuard(),CodeNode.voidActions());
				m.put(s, t);
			}
			break;
		case Statement.while_: 
			i = instruction.next(heap);
			if(i == Statement.double_)
			{
				// 2 branchement possibles
				s = new State();
				s.heap = heap;
				s.instruction = instruction.next(true);
				s.isFinal = (s.instruction == null)?true:false;
				CodeNode a = (CodeNode)((CodeNode)instruction).jjtGetChild(0);
				t = new Transition(this,s,pointerGuards?a:CodeNode.voidGuard(),CodeNode.voidActions());
				m.put(s, t);
				
				s = new State();
				s.heap = heap;
				s.instruction = instruction.next(false);
				s.isFinal = (s.instruction == null)?true:false;
				CodeNode g = a.isVoid()?a:CodeNode.not(a);
				t = new Transition(this,s,pointerGuards?g:CodeNode.voidGuard(),CodeNode.voidActions());
				m.put(s, t);				
			}
			else
			{
				// un seul branchement possible
				s = new State();
				s.heap = heap;
				s.instruction = instruction.next((i == Statement.true_)?true:false);
				s.isFinal = (s.instruction == null)?true:false;
				CodeNode a = (CodeNode)((CodeNode)instruction).jjtGetChild(0);
				CodeNode g = (i == Statement.true_ || a.isVoid())?a:CodeNode.not(a);
				t = new Transition(this,s,pointerGuards?g:CodeNode.voidGuard(),CodeNode.voidActions());
				m.put(s, t);
			}
			break;
		case Statement.program_: 
			s = new State();
			s.heap = heap;
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.voidActions());
			m.put(s, t);
			break;
		case Statement.slist_: 
			s = new State();
			s.heap = heap;
			s.instruction = instruction.next();
			s.isFinal = (s.instruction == null)?true:false;
			
			t = new Transition(this,s,CodeNode.voidGuard(),CodeNode.voidActions());
			m.put(s, t);
			break;
		}
		return m;
	}

	public String toString() {
		return heap.toString() + instruction.toString()+" \n " + isFinal() + "\n";
	}

	public int hashCode() {
		final int PRIME = 31;
		int result = 1;
		result = PRIME * result + ((heap == null) ? 0 : heap.hashCode());
		result = PRIME * result + id;
		result = PRIME * result + (isFinal ? 1231 : 1237);
		return result;
	}

	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final State other = (State) obj;
		if (instruction == null) {
			if (other.instruction != null)
				return false;
		} else if (!instruction.equals(other.instruction))
			return false;
		if (heap == null) {
			if (other.heap != null)
				return false;
		} else if (!heap.equals(other.heap))
			return false;		
		if (isFinal != other.isFinal)
			return false;
		return true;
	}
	
	public Object isEquivalentTo(Object obj)
	{
		if (this == obj)
			return Equivalent.Identity;
		if (obj == null)
			return null;
		if (getClass() != obj.getClass())
			return null;
		final State other = (State) obj;
		if (instruction == null) {
			if (other.instruction != null)
				return null;
		} else if (!instruction.equals(other.instruction))
			return null;
		if (heap == null) {
			if (other.heap != null)
				return null;
		} else if (heap.isEquivalentTo(other.heap) == null)
			return null;		
		if (isFinal != other.isFinal)
			return null;
		return Equivalent.Identity;
	}	
	
	
}
