package output;

import java.io.FileNotFoundException;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import code.CodeNode;

import automaton.*;

public class ArmcFile implements FileOutput {
	private java.io.PrintWriter file;
	private String name;
	private boolean normal = true;
	private String data,data1;
	
	public ArmcFile()
	{
		file = null;
		name = null;
	}
	
	public String add() {
		return " + ";
	}

	public String and() {
		if(normal)
			return " , ";
		else
			return " | ";
	}

	public String asgn() {
		return "1 = ";
	}

	public void close() {
		file.flush();
		file.close();
	}

	public String eq() {
		if(normal)
			return " = ";
		else
			return " =\\= ";
	}

	public String ge() {
		if(normal)
			return " >= ";
		else
			return " < ";
	}

	public String gt() {
		if(normal)
			return " > ";
		else
			return " <= ";
	}

	public String le() {
		if(normal)
			return " <= ";
		else
			return " > ";
	}

	public String lt() {
		if(normal)
			return " < ";
		else
			return " >= ";
	}

	public String mult() {
		return " * ";
	}

	public String not() {
		return "";
	}

	public void open(String name) throws FileNotFoundException {
		this.name = name;
		file = new java.io.PrintWriter(new java.io.FileOutputStream(name+".armc"));
	}

	public String filename(){
		return name+".armc";
	}
	
	public String or() {
		if(normal)
			return " | ";
		else
			return " , ";
	}

	public String slist() {
		return " , ";
	}

	public String sub() {
		return " - ";
	}
	
	public String Null(){
		return " null ";
	}
	
	public String New(){
		return " new ";
	}

	public void switchNot() {
		normal = !normal;
	}

	public void write(Automaton a) {
		
		data = this.firstData(a);
		data1 = this.secondData(a);
		
		// init 
		this.file.println(":- multifile r/5, implicit_updates/0, var2names/2, preds/2, cube_size/1, start/1, refinement/1, error/1.");
		this.file.println("refinement(inter).\ncube_size(1).\nimplicit_updates.\n");


		// start 
		this.file.println("start(pc(h_"+a.getFirst().getId()+")).");
		this.file.println("error(pc(h_error)).");
		

		// var2names 
		boolean first = true;
		this.file.println("var2names(p(_,"+data+"),[");
		Iterator i = a.getCounters().iterator();
		while(i.hasNext())
		{
			String c = (String)i.next();
			if(first)
			{
			    this.file.print("("+c+",'"+c+"')");
			    first = false;
			}
			else
			{
			    this.file.print(",("+c+",'"+c+"')");
			}
		}
		this.file.println("]).");

		// preds 
		this.file.println("preds(p(_,"+data+"),[]).");


		// transitions  
		i = a.getTransitions().iterator();
		int id = 0;
		while(i.hasNext())
		{
		    this.file.print("r(");
		    this.Writetransition((Transition)i.next(), id);
		    this.file.println(").");
		    id++;
		}		
	}
	
	private void Writetransition(Transition t,int id)
	{
		this.file.print("p(");
		if(t.getSinitial().isError())
			this.file.print("pc(h_error),"+data+"),");
		else
			this.file.print("pc(h_"+t.getSinitial().getId()+"),"+data+"),");
		
		this.file.print("p(");
		if(t.getSfinal().isError())
			this.file.print("pc(h_error),"+data1+"),");
		else
			this.file.print("pc(h_"+t.getSfinal().getId()+"),"+data1+"),");
		
		this.file.print("["+this.writeGuard(t.getGuard())+"],");
		this.file.print("["+this.writeActions(t.getActions())+"],");
		this.file.print(id);
	}
	
	private String writeGuard(CodeNode g)
	{
		if(g == null)
			return " ";
		else
			return g.out(this);
	}
	
	private String writeActions(CodeNode a)
	{
		if(a == null)
			return " ";
		else
			return a.out(this);
	}

	private String secondData(Automaton a) {
		String s = "data(";
		Iterator i = a.getCounters().iterator();
		boolean first = true;
		while(i.hasNext())
		{
			if(first)
			{
				s += (String) i.next();
				first = false;
			}
			else
			{
				s += "," + (String) i.next() + "1";
			}
		}
		s += ")";
		return s;
	}

	private String firstData(Automaton a) {
		String s = "data(";
		Iterator i = a.getCounters().iterator();
		boolean first = true;
		while(i.hasNext())
		{
			if(first)
			{
				s += (String) i.next();
				first = false;
			}
			else
			{
				s += "," + (String) i.next();
			}
		}
		s += ")";
		return s;
	}

		
	public boolean pointerGuards()
	{
		return false;
	}

	public String leftP() {
		return " ( ";
	}

	public String rightP() {
		return " ) ";
	}

	public boolean supportSwitch() {
		return true;
	}

}
