package output;
import java.util.*;
import automaton.*;
import heap.*;
import code.*;
import util.*;

public class DotFile implements FileOutput {
	private java.io.PrintWriter file;
	private String name;
	private String filename;
	private boolean normal = true;
	
	public DotFile()
	{
		file = null;
		name = null;
		filename = null;
	}
	
	
	public void close() {
		file.flush();
		file.close();
	}

	public void open(String name) throws java.io.FileNotFoundException {
		this.name = name;
		int i = this.name.indexOf('/');
		if(i != -1)
			this.name = this.name.substring(i+1);
		
		file = new java.io.PrintWriter(new java.io.FileOutputStream(name+".dot"));
		this.filename = name + ".dot"; 
	}
	
	public String filename(){
		return filename;
	}

	public void write(Automaton a) {
		file.println("digraph " + name + " {");
		this.autStyle();
		/*int p = 1;
		MyPriorityQueue q = rankedStates(a.getStates());
		while(!q.isEmpty())
		{
			Element e = (Element)q.get();
			State s = (State) e.getElement();
			int priority = e.getPriority();
			s.setId(p);
			if(s.isError())
			{
				file.println("\t subgraph clusterERROR {");
				file.println("\t style=dashed;");
				this.heapStyle();
				file.println(" ERROR ");
				file.println(" }");
			}
			else
			{
				Heap h = s.getHeap();
				file.println("\t subgraph cluster" + s.getId() +" {");
				file.println("\t style=dashed;");
				this.heapStyle();
				file.println("\t \t n_"+s.getId()+"_"+priority+"[label=\""+priority+"\"];");
				this.heapWrite(h,s.getId());
				file.println(" }");
			}
			p++;
		}*/		
		Iterator i = a.getStates().iterator();
		while(i.hasNext())
		{
			State s = (State) i.next();
			if(s.isError())
			{
				file.println("\t subgraph clusterERROR {");
				file.println("\t style=dashed;");
				this.heapStyle();
				file.println(" ERROR ");
				file.println(" }");
			}
			else
			{
				Heap h = s.getHeap();
				file.println("\t subgraph cluster" + s.getId() +" {");
				file.println("\t style=dashed;");
				this.heapStyle();
				this.heapWrite(h,s.getId());
				file.println(" }");
			}
		}
		i = a.getTransitions().iterator();
		while(i.hasNext())
		{
			Transition t = (Transition)i.next();
			int cluster_id_is = t.getSinitial().getId();
			int cluster_id_fs = t.getSfinal().getId();
			if(cluster_id_is == cluster_id_fs)
			{
				int is_first_v,fs_first_v;
				if(t.getSinitial().getHeap().getVertices().size() != 0)
					is_first_v = ((Vertex)t.getSinitial().getHeap().getVertices().get(0)).getId();
				else
					is_first_v = 0;
				fs_first_v = is_first_v;
				file.println(" h_"+cluster_id_is+"_"+is_first_v+" -> h_"+cluster_id_fs+"_"+fs_first_v+" [label=\""+this.guardWrite(t.getGuard())+" "+this.actionWrite(t.getActions())+"\",minlen=1.5] ;");
			}
			else if(!t.getSinitial().isError() && !t.getSfinal().isError())
			{
				int is_first_v,fs_first_v;
				if(t.getSinitial().getHeap().getVertices().size() != 0)
					is_first_v = ((Vertex)t.getSinitial().getHeap().getVertices().get(0)).getId();
				else
					is_first_v = 0;
				if(t.getSfinal().getHeap().getVertices().size() != 0)
					fs_first_v = ((Vertex)t.getSfinal().getHeap().getVertices().get(0)).getId();
				else
					fs_first_v = 0;
				
				file.println(" h_"+cluster_id_is+"_"+is_first_v+" -> h_"+cluster_id_fs+"_"+fs_first_v+" [ltail=cluster"+cluster_id_is+" ,lhead=cluster"+cluster_id_fs+" ,label=\""+this.guardWrite(t.getGuard())+" "+this.actionWrite(t.getActions())+"\",minlen=1.5] ;");	
			}
			else
			{
				int is_first_v,fs_first_v;
				if(t.getSinitial().isError())
				{
					
					if(t.getSfinal().getHeap().getVertices().size() != 0)
						fs_first_v = ((Vertex)t.getSfinal().getHeap().getVertices().get(0)).getId();
					else
						fs_first_v = 0;
					
					file.println(" ERROR -> h_"+cluster_id_fs+"_"+fs_first_v+" [ltail=clusterERROR ,lhead=cluster"+cluster_id_fs+" ,minlen=1.5] ;");
				}
				else
				{
					if(t.getSinitial().getHeap().getVertices().size() != 0)
						is_first_v = ((Vertex)t.getSinitial().getHeap().getVertices().get(0)).getId();
					else
						is_first_v = 0;
					
					file.println("h_"+cluster_id_is+"_"+is_first_v+" -> ERROR  [ltail=cluster"+cluster_id_is+" ,lhead=clusterERROR ,minlen=1.5] ;");
				}					
			}
		}		
		file.println("}");
		file.flush();
	}

	private void heapStyle(){
		file.println("node [shape=box,regular=true];");
	}
	
	private void heapWrite(Heap h,int id){
		Iterator i = h.getVertices().iterator();
		int j = 0;
		while(i.hasNext())
		{
			Vertex v = (Vertex)i.next();
			file.print("h_"+id+"_"+v.getId()+" [label=\"");
			Iterator k = h.getRoots(v).iterator();
			while(k.hasNext())
			{
				Root r = (Root)k.next();
				file.print(r.name()+" ");
			}
			file.println(" \\n "+v.counter()+"\"];");
			j++;
		}
		if(j == 0)
			file.println("h_"+id+"_"+j+" [label=\"Void\"];");
		Iterator l = h.getVertices().iterator();
		while(l.hasNext())
		{
			Vertex v = (Vertex)l.next();
			if(h.getSuccessor(v) != null)
			file.println("h_"+id+"_"+v.getId()+" -> h_"+id+"_"+h.getSuccessor(v).getId()+" [minlen=0];");
		}		
	}
	
	private void autStyle(){
		file.println("node [shape=box,regular=false];");
	}
	
	private String guardWrite(CodeNode n){
		if(n == null)
			return " ";
		else
			return "[" + n.out(this) + "]\\n";
	}
	
	private String actionWrite(CodeNode n){
		if(n == null)
			return " ";
		else
			return n.out(this);
	}
	
	private MyPriorityQueue rankedStates(List states)
	{
		MyPriorityQueue p = new MyPriorityQueue();
		Iterator i = states.iterator();
		int max = 0;
		while(i.hasNext())
		{
			State s = (State)i.next();
			CodeNode n = (CodeNode)s.getInstruction();
			if( n != null)
				if(n.getNumber() > max)
					max = n.getNumber();
		}
		max++;
		i = states.iterator();
		while(i.hasNext())
		{
			State s = (State)i.next();
			CodeNode n = (CodeNode)s.getInstruction();
			int e;
			if(n == null)
			{
				if(s.isFinal())
					e = max;
				else
					e = 0;
			}
			else
				e = n.getNumber();
			p.add(s,e);
		}
		return p;
	}

	public String add() {
		return " + ";
	}

	public String and() {
		if(normal)
			return " & ";
		else
			return " | ";
	}


	public String asgn() {
		return " := ";
	}


	public String eq() {
		if(normal)
			return " = ";
		else
			return " != ";
	}


	public String ge() {
		if(normal)
			return " >= ";
		else
			return " < ";
	}


	public String gt() {
		if(normal)
			return " > ";
		else
			return " <= ";
	}


	public String le() {
		if(normal)
			return " <= ";
		else
			return " > ";
	}


	public String lt() {
		if(normal)
			return " < ";
		else
			return " >= ";
	}


	public String mult() {
		return " * ";
	}


	public String not() {
		if(normal)
			return " ! ";
		else
			return " ";
	}


	public String or() {
		if(normal)
			return " | ";
		else
			return " & ";
	}


	public String slist() {
		return " ;\\n ";
	}


	public String sub() {
		return " - ";
	}
	
	public String Null(){
		return " null ";
	}
	
	public String New(){
		return " new ";
	}
	
	public boolean pointerGuards()
	{
		return true;
	}
	
	public void switchNot(){
		normal = !normal;
	}


	public String leftP() {
		return " ( ";
	}


	public String rightP() {
		return " ) ";
	}


	public boolean supportSwitch() {
		return true;		
	}
}
